# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/runtime_env_agent.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import agent_manager_pb2 as src_dot_ray_dot_protobuf_dot_agent__manager__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='src/ray/protobuf/runtime_env_agent.proto',
  package='ray.rpc',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n(src/ray/protobuf/runtime_env_agent.proto\x12\x07ray.rpc\x1a$src/ray/protobuf/agent_manager.proto\"f\n\x17\x43reateRuntimeEnvRequest\x12\x34\n\x16serialized_runtime_env\x18\x01 \x01(\tR\x14serializedRuntimeEnv\x12\x15\n\x06job_id\x18\x02 \x01(\x0cR\x05jobId\"\xb2\x01\n\x15\x43reateRuntimeEnvReply\x12/\n\x06status\x18\x01 \x01(\x0e\x32\x17.ray.rpc.AgentRpcStatusR\x06status\x12#\n\rerror_message\x18\x02 \x01(\tR\x0c\x65rrorMessage\x12\x43\n\x1eserialized_runtime_env_context\x18\x03 \x01(\tR\x1bserializedRuntimeEnvContext\"O\n\x17\x44\x65leteRuntimeEnvRequest\x12\x34\n\x16serialized_runtime_env\x18\x01 \x01(\tR\x14serializedRuntimeEnv\"m\n\x15\x44\x65leteRuntimeEnvReply\x12/\n\x06status\x18\x01 \x01(\x0e\x32\x17.ray.rpc.AgentRpcStatusR\x06status\x12#\n\rerror_message\x18\x02 \x01(\tR\x0c\x65rrorMessage2\xbf\x01\n\x11RuntimeEnvService\x12T\n\x10\x43reateRuntimeEnv\x12 .ray.rpc.CreateRuntimeEnvRequest\x1a\x1e.ray.rpc.CreateRuntimeEnvReply\x12T\n\x10\x44\x65leteRuntimeEnv\x12 .ray.rpc.DeleteRuntimeEnvRequest\x1a\x1e.ray.rpc.DeleteRuntimeEnvReplyb\x06proto3'
  ,
  dependencies=[src_dot_ray_dot_protobuf_dot_agent__manager__pb2.DESCRIPTOR,])




_CREATERUNTIMEENVREQUEST = _descriptor.Descriptor(
  name='CreateRuntimeEnvRequest',
  full_name='ray.rpc.CreateRuntimeEnvRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='serialized_runtime_env', full_name='ray.rpc.CreateRuntimeEnvRequest.serialized_runtime_env', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='serializedRuntimeEnv', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='job_id', full_name='ray.rpc.CreateRuntimeEnvRequest.job_id', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='jobId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=91,
  serialized_end=193,
)


_CREATERUNTIMEENVREPLY = _descriptor.Descriptor(
  name='CreateRuntimeEnvReply',
  full_name='ray.rpc.CreateRuntimeEnvReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='ray.rpc.CreateRuntimeEnvReply.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='status', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error_message', full_name='ray.rpc.CreateRuntimeEnvReply.error_message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='errorMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='serialized_runtime_env_context', full_name='ray.rpc.CreateRuntimeEnvReply.serialized_runtime_env_context', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='serializedRuntimeEnvContext', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=196,
  serialized_end=374,
)


_DELETERUNTIMEENVREQUEST = _descriptor.Descriptor(
  name='DeleteRuntimeEnvRequest',
  full_name='ray.rpc.DeleteRuntimeEnvRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='serialized_runtime_env', full_name='ray.rpc.DeleteRuntimeEnvRequest.serialized_runtime_env', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='serializedRuntimeEnv', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=376,
  serialized_end=455,
)


_DELETERUNTIMEENVREPLY = _descriptor.Descriptor(
  name='DeleteRuntimeEnvReply',
  full_name='ray.rpc.DeleteRuntimeEnvReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='ray.rpc.DeleteRuntimeEnvReply.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='status', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error_message', full_name='ray.rpc.DeleteRuntimeEnvReply.error_message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='errorMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=457,
  serialized_end=566,
)

_CREATERUNTIMEENVREPLY.fields_by_name['status'].enum_type = src_dot_ray_dot_protobuf_dot_agent__manager__pb2._AGENTRPCSTATUS
_DELETERUNTIMEENVREPLY.fields_by_name['status'].enum_type = src_dot_ray_dot_protobuf_dot_agent__manager__pb2._AGENTRPCSTATUS
DESCRIPTOR.message_types_by_name['CreateRuntimeEnvRequest'] = _CREATERUNTIMEENVREQUEST
DESCRIPTOR.message_types_by_name['CreateRuntimeEnvReply'] = _CREATERUNTIMEENVREPLY
DESCRIPTOR.message_types_by_name['DeleteRuntimeEnvRequest'] = _DELETERUNTIMEENVREQUEST
DESCRIPTOR.message_types_by_name['DeleteRuntimeEnvReply'] = _DELETERUNTIMEENVREPLY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CreateRuntimeEnvRequest = _reflection.GeneratedProtocolMessageType('CreateRuntimeEnvRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATERUNTIMEENVREQUEST,
  '__module__' : 'src.ray.protobuf.runtime_env_agent_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.CreateRuntimeEnvRequest)
  })
_sym_db.RegisterMessage(CreateRuntimeEnvRequest)

CreateRuntimeEnvReply = _reflection.GeneratedProtocolMessageType('CreateRuntimeEnvReply', (_message.Message,), {
  'DESCRIPTOR' : _CREATERUNTIMEENVREPLY,
  '__module__' : 'src.ray.protobuf.runtime_env_agent_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.CreateRuntimeEnvReply)
  })
_sym_db.RegisterMessage(CreateRuntimeEnvReply)

DeleteRuntimeEnvRequest = _reflection.GeneratedProtocolMessageType('DeleteRuntimeEnvRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETERUNTIMEENVREQUEST,
  '__module__' : 'src.ray.protobuf.runtime_env_agent_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.DeleteRuntimeEnvRequest)
  })
_sym_db.RegisterMessage(DeleteRuntimeEnvRequest)

DeleteRuntimeEnvReply = _reflection.GeneratedProtocolMessageType('DeleteRuntimeEnvReply', (_message.Message,), {
  'DESCRIPTOR' : _DELETERUNTIMEENVREPLY,
  '__module__' : 'src.ray.protobuf.runtime_env_agent_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.DeleteRuntimeEnvReply)
  })
_sym_db.RegisterMessage(DeleteRuntimeEnvReply)



_RUNTIMEENVSERVICE = _descriptor.ServiceDescriptor(
  name='RuntimeEnvService',
  full_name='ray.rpc.RuntimeEnvService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=569,
  serialized_end=760,
  methods=[
  _descriptor.MethodDescriptor(
    name='CreateRuntimeEnv',
    full_name='ray.rpc.RuntimeEnvService.CreateRuntimeEnv',
    index=0,
    containing_service=None,
    input_type=_CREATERUNTIMEENVREQUEST,
    output_type=_CREATERUNTIMEENVREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteRuntimeEnv',
    full_name='ray.rpc.RuntimeEnvService.DeleteRuntimeEnv',
    index=1,
    containing_service=None,
    input_type=_DELETERUNTIMEENVREQUEST,
    output_type=_DELETERUNTIMEENVREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_RUNTIMEENVSERVICE)

DESCRIPTOR.services_by_name['RuntimeEnvService'] = _RUNTIMEENVSERVICE

# @@protoc_insertion_point(module_scope)
